use futures::Future;
use hyper::rt::Executor;
use hyper::service::{make_service_fn, service_fn};
use hyper::{Body, Response, Server};
use std::convert::Infallible;
use std::net::SocketAddr;
use std::sync::Arc;
use tokio::sync::Mutex;
use tracing::error;
use zerocopy::{AsBytes, FromBytes, LittleEndian, Unaligned};

use tracing::{debug, instrument};

use crate::packets::{Header, Packet};

use crate::ports::PortHandler;
use crate::spawn;

#[derive(Clone)]
struct NamedExecutor;
impl<T: Send + 'static, Fut: Future<Output = T> + Send + 'static> Executor<Fut> for NamedExecutor {
    fn execute(&self, fut: Fut) {
        spawn("http worker", fut);
    }
}

pub async fn debug_server(addr: SocketAddr, port_handler: Arc<Mutex<PortHandler>>) {
    let server = Server::bind(&addr)
        .executor(NamedExecutor)
        .serve(make_service_fn(move |_conn| {
            let port_handler = port_handler.clone();
            async move {
                Ok::<_, Infallible>(service_fn(move |_req| {
                    let port_handler = port_handler.clone();
                    async move {
                        Ok::<_, Infallible>(Response::new(Body::from(
                            port_handler.lock().await.status_string(),
                        )))
                    }
                }))
            }
        }));

    // Run this server for... forever!
    if let Err(error) = server.await {
        error!(%error, "debug server error");
    }
}

type U16 = zerocopy::U16<LittleEndian>;
type U32 = zerocopy::U32<LittleEndian>;

#[derive(AsBytes)]
#[repr(transparent)]
#[allow(dead_code)]
struct PeerQuery {
    number: U32,
}

#[derive(FromBytes, Unaligned, Debug)]
#[repr(packed)]
#[allow(dead_code)]
struct PeerReply {
    number: U32,
    name: [u8; 40],
    flags: U16,
    kind: u8,
    hostname: [u8; 40],
    ipaddress: [u8; 4],
    port: U16,
    extension: u8,
    pin: U16,
    timestamp: U32,
}

#[instrument]
pub async fn peer_query(server: &SocketAddr, number: u32) -> eyre::Result<Option<String>> {
    debug!(%number, "looking up");

    let mut packet = Packet {
        header: Header {
            kind: 3, // Peer Query
            length: 4,
        },
        data: Vec::new(),
    };

    packet.data.clear();
    packet.data.resize(packet.header.length as usize, 0);

    PeerQuery {
        number: number.into(),
    }
    .write_to(packet.data.as_mut_slice())
    .unwrap();

    let mut socket = tokio::net::TcpStream::connect(server).await?;

    let (mut reader, mut writer) = socket.split();

    packet.send(&mut writer).await?;
    packet.recv_into(&mut reader).await?;

    Ok(if packet.kind().raw() == 5 {
        // PeerReply
        PeerReply::read_from(packet.data.as_slice()).and_then(|x| {
            let i = x
                .name
                .iter()
                .enumerate()
                .find_map(|(i, c)| (*c == 0).then_some(i))
                .unwrap_or(x.name.len());

            Some(std::str::from_utf8(&x.name[..i]).ok()?.to_owned())
        })
    } else {
        None
    })
}
